<?php

/*
Plugin Name: Plugin Profiler
Plugin URI: http://w-shadow.com/
Description: Shows how much time is spent running each of your plugins on any particular page. This lets you find slow plugins that can negatively impact your site's performance. 
Version: 1.0
Author: Janis Elsts
Author URI: http://w-shadow.com/blog/2008/09/23/wordpress-plugin-profiler/

Compatibility (min) :
 	PHP 5.0.0 (uses microtime(true) because I'm lazy)
	WP 2.6, tested in 2.6.2 (patch assumes _wp_call_all_hook is present)  
*/

/**
 * @author W-Shadow
 * @copyright 2008
 */

if (!class_exists('PluginProfiler')) {
	
class PluginProfiler {
	var $options_name = 'plugin_profiler';
	var $patches = array(
		'wp-db.php' => 'wp-includes/wp-db.php',
		'plugin.php' => 'wp-includes/plugin.php',
		'wp-settings.php' => 'wp-settings.php',
	);
	
 
	function PluginProfiler(){
		//Set up option defaults
		$this->defaults = array(
			'output' => 'none', //Possible values : none, plaintext, plaintext_hidden, advanced
			'output_admin_only' => true, //Only show the output to logged-in admin users
		);
		
		//Load options
		$this->options = get_option($this->options_name);
		
		//Use defaults when options are not set
		if(!is_array($this->options)){
			$this->options = $this->defaults;				
		} else {
			$this->options = array_merge($this->defaults, $this->options);
		}
		
		//add_action('admin_footer', array(&$this, 'admin_footer'));
		add_action('shutdown', array(&$this, 'shutdown_hook'), 1000);
		add_action('admin_menu', array(&$this, 'add_admin_menus'));
		add_action('wp_footer', array(&$this, 'wp_footer'));
		
		if ( $this->options['output'] == 'advanced'  ) {
			add_action('admin_print_scripts', array(&$this,'enqueue_scripts'));
			
			if (!$this->options['output_admin_only']){
				add_action('wp_print_scripts', array(&$this,'enqueue_scripts'));
			}
		} 
		
		$this->my_basename = plugin_basename(__FILE__);
	}
	
  /**
   * PluginProfiler::admin_footer()
   * Handler for the admin_footer hook. Purely for debuggin purposes.
   *
   * @return void
   */
	function admin_footer(){
		$trace = debug_backtrace();
		echo "<pre>", print_r($trace, true), "<pre>";
		
	}
	
	function wp_footer(){
		//This function does nothing. It's here just so that the incremental time of wp_footer can be recorded.
	}
	
  /**
   * PluginProfiler::add_admin_menus()
   * Handler for the add_admin_menus hook. Adds the settings menu and sets a hook 
   * for adding a direct "Settings" link to the plugin's action links.
   *
   * @return void
   */
	function add_admin_menus(){
		add_submenu_page('options-general.php', 'Plugin Profiler', 'Plugin Profiler', 'edit_plugins', 
			'plugin_profiler_settings', array(&$this, 'options_page'));
		
		if (current_user_can('edit_plugins')) 
			add_filter('plugin_action_links', array(&$this, 'plugin_action_links'), 10, 2);
	}
	
  /**
   * PluginProfiler::plugin_action_links()
   * Adds a "Settings" link to the plugin's action links
   *
   * @param array $links
   * @param string $file
   * @return array
   */
	function plugin_action_links($links, $file){
		if ($file == $this->my_basename)
			$links[] = "<a href='options-general.php?page=plugin_profiler_settings'>" . __('Settings') . "</a>";
		return $links;
	}
	
  /**
   * PluginProfiler::enqueue_scripts()
   * Tells WP to load jQuery 
   *
   * @return void
   */
	function enqueue_scripts(){
   		wp_enqueue_script('jquery');
	}
	
  /**
   * PluginProfiler::shutdown_hook()
   * Outputs the profiler report after the rest of the page has already been generated.
   * Handles the "shutdown" hook.
   *
   * @return void
   */
	function shutdown_hook(){
		global $plugin_stats, $hook_stats;
		
		//Skip output if user has opted to only show it to logged-in admins and the current visitor isn't one.
		if ( $this->options['output_admin_only'] && !current_user_can('edit_plugins') ) return;
		if ( $this->options['output']=='none' ) return;
		
		
		if (!isset($plugin_stats) || !is_array($plugin_stats)){
			echo "<div style='clear:both;'> </div>
			The profiler is running, but no stats are available! Most likely you haven't
			installed the required patch(es).";
			return;
		}
		
		//Output execution stats (for the current page) in the user's chosen format
		switch($this->options['output']){
			case 'plaintext': //Very simple, just a call to print_r
				$this->print_simple_stats();
				break;
				
			case 'plaintext_hidden': //As above, but in a HTML comment
				echo "<!--\n";
				$this->print_simple_stats();
				echo "\n-->";
				break;
				
			case 'advanced': //Nice output with sorting and other features
				$this->print_advanced_stats();
				break; 
		}		
	}
	
	
	function can_remove_patch() {
		foreach ($this->patches as $patch => $target){
			if (!file_exists(ABSPATH.$target.'.backup'))
				return false;
			if (!is_writable(ABSPATH.$target))
				return false;
		}
		return true;
	}
	
	function can_patch() {
		global $wp_version;
		$patch_dir = dirname(__FILE__).'/patches/'.$wp_version.'/';
		
		foreach ($this->patches as $patch => $target){
			if (!file_exists($patch_dir.$patch))
				return false;
			if (!is_writable(ABSPATH.$target))
				return false;
		}
		return true;
	}
	
  /**
   * PluginProfiler::install_patch()
   * Installs the patched version of plugins.php. Doesn't check file permissions!
   *
   * @return boolean
   */
	function install_patch(){
		global $wp_version;
		
		//Ensure I don't patch an already patched file.
		if (function_exists('_ws_record_execution_time')) return false;
		
		
		$patch_dir = dirname(__FILE__).'/patches/'.$wp_version.'/';
		
		//Apply all patches
		foreach ($this->patches as $patch => $target){
			$backup_file = ABSPATH.$target.'.backup';
			$target = ABSPATH.$target;
			$patch = $patch_dir.$patch;
			//This should never happen... Right, like that ever works :)
			if (file_exists($backup_file)){
				unlink($backup_file);
			}
			
			//Save the original as a backup
			if (!rename($target, $backup_file)) return false;
			
			//Copy the patch in its place
			if (!copy($patch, $target)) return false;
				
		}
		
		return true;
	}
	
  /**
   * PluginProfiler::remove_patch()
   * Removes the patched plugin.php file and restores the original. Doesn't check file permissions!
   *
   * @return boolean
   */
	function remove_patch(){
		//Ensure the patch is actually present.
		if (!function_exists('_ws_record_execution_time')) return false;
		
		//Apply all patches
		foreach ($this->patches as $patch => $target){
			$backup_file = ABSPATH.$target.'.backup';
			$target = ABSPATH.$target;
			
			//Remove the patched file
			if (!unlink($target)) return false;
			
			//Restore it from backup
			if (!rename($backup_file, $target)) return false;
			
		}
		
		return true;
	}
	
  /**
   * PluginProfiler::options_page()
   * Displays the settings page for the plugin.
   *
   * @return void
   */
	function options_page(){
		global $wp_version;
		//print_r($_POST);
		
		if (!empty($_POST['install_patch']) ) {
			if ($this->install_patch()){
?>
			<div class='updated'><h3>Success!</h3>
			<p>Some core files were replaced with patched versions. The original files are still 
			available as <code>backup.[filename]</code>. <br />
			These files were patched : 
			<ul>
				<?php
					foreach ($this->patches as $f) echo "<li>$f</li>";
				?>
			</ul>  
			</p>
			<p><a href='options-general.php?page=plugin_profiler_settings'>Continue to Settings</a></p>
			</div>
<?php
			} else {
?>
			<div class='error'><h3>Error</h3>
			<p>The patch couldn't be installed. The file permissions seemed OK and 
			it looked like it would work, but for some reason it didn't. Oops.
			</p>
			<p><a href='options-general.php?page=plugin_profiler_settings'>Back to Settings</a></p>
			</div>
<?php
			};
			return;
		} else if (!empty($_POST['remove_patch']) ) {
			if ($this->remove_patch()){
?>
			<div class='updated'><h3>Success!</h3>
			<p>The original files were successfully restored from backups.
			
			The patch has been uninstalled.</p>
			<p><a href='options-general.php?page=plugin_profiler_settings'>Continue to Settings</a></p>
			</div>
<?php
			} else {
?>
			<div class='error'><h3>Error</h3>
			<p>The patch couldn't be removed. The file permissions seemed OK and 
			it looked like it would work, but for some reason it didn't. Oops.
			</p>
			<p><a href='options-general.php?page=plugin_profiler_settings'>Back to Settings</a></p>
			</div>
<?php
			};
			return;	
		} else if (!empty($_POST['action']) && ($_POST['action']=='update') ){
			$this->options['output'] = $_POST['output'];
			$this->options['output_admin_only'] = !empty($_POST['output_admin_only']);
			
			update_option($this->options_name, $this->options);

			echo '<div id="message" class="updated fade"><p><strong>Settings saved.</strong></p></div>';
		} 
		
		?>
<div class="wrap">
<h2>Plugin Profiler Settings</h2>

<form name="plugin_monitor_patch" method="post" 
action="<?php echo $_SERVER['PHP_SELF']; ?>?page=plugin_profiler_settings">

<h3>Patch Status</h3>
<table class="form-table">
	<tr>
		<td colspan="2" style='font-size: 10pt;'>
		<!-- The *tiny* fonts that WP 2.6 uses by default really bother me. -->
<?php
	$patch_installed = function_exists('_ws_record_execution_time');
	if ($patch_installed){
	/** ---------- Patch is already installed ---------- */
		
?>
	<strong>Patch installed.</strong><br /> 
	The profiler patch is active. Note that you will need to remove the patch to 
	disable monitoring. If you only deactivate or delete this plugin the patch will continue working 
	(but no output will be produced).
<?php
	//Check if the patch can be automatically removed.
	if ( $this->can_remove_patch() ){
?>
	<input type='submit' name='remove_patch' value='Remove patch' />
<?php		
	} else {
		echo "If you wish to remove the patch, replace ",
			join(', ', array_keys($this->patches)),	
		     " with the original files that you have backed up.";
	}
?>  
<?php		
	} else {
	/** ---------- Patch is not installed ---------- */
?>
	<strong class='error'>Patch not installed.</strong> <br />
 	Several core files need to be replaced with patched versions to enable plugin profiling. 
<?php

		//See if there is a patch available for this version of WP
		$patch_available = file_exists(dirname(__FILE__).'/patches/'.$wp_version.'/plugin.php');
			
		if ($patch_available){
			//echo "Patch is available. ";
			if ( $this->can_patch() ){
?>
Click "Install patch" to let the plugin do this. The original files will be automatically backed up.
<br />  
	<input type='submit' name='install_patch' value='Install patch' />
<?
			} else {
?>
			To install the patch, you need to manually replace these files :  
			<ul>
			<?php 
			$patch_dir = basename(dirname(__FILE__)).'/patches/'.$wp_version.'/';
			foreach ($this->patches as $patch => $target){
				echo "<li><code>$target</code> with {$patch_dir}{$patch}<li>";
			}
			?>
			</ul>
			
			Please backup your WordPress files beforehand - you will need the original files to remove the patch.  
<?php				
			}
		} else {
			echo "There is no patch available for WordPress $wp_version! You can't use the plugin
			with this version of WordPress.";
		}

	}
?>		
		<td>
	</tr>
</table>
</form>

<form name="plugin_monitor_options" method="post" 
action="<?php echo $_SERVER['PHP_SELF']; ?>?page=plugin_profiler_settings">

<input type='hidden' name='action' value='update' />

<h3>Output Settings</h3>
<table class="form-table" style='font-size: 10pt;'>
	<tr>
		<th scope="row">Output Format</th>
		<td style='font-size: 10pt;'>
		<fieldset>
		
		<p><label><input name="output" type="radio" value="none" 
		<?php if ($this->options['output']=='none') echo "checked='checked'"; ?> /> 
			<b>None.</b> This turns off the display of statistics.   
		</label></p>
		
		<p><label><input name="output" type="radio" value="advanced" 
		<?php if ($this->options['output']=='advanced') echo "checked='checked'"; ?> /> 
			<b>Advanced.</b> Displays stats in a neatly formatted and sorted manner with 
				several navigation options. Recommended.
		</label></p>
		
		<p><label><input name="output" type="radio" value="plaintext" 
		<?php if ($this->options['output']=='plaintext') echo "checked='checked'"; ?> /> 
			<b>Plain text.</b> Execution statistics are displayed in a very simple, purely textual format. 
				Basically this is just <code>print_r</code>. Not recommended for most users.  
		</label></p>
		
		<p><label><input name="output" type="radio" value="plaintext_hidden" 
		<?php if ($this->options['output']=='plaintext_hidden') echo "checked='checked'"; ?> /> 
			<b>Plain text in a HTML comment.</b> As above, but the output is hidden in a HTML comment.  
		</label></p>
		
		</fieldset>
		</td>
	</tr>
	
	<tr>
		<th align='left'>Admin-only</th>
		<td style='font-size: 10pt;'>
		<label><input type='checkbox' name='output_admin_only' id='output_admin_only' <?php
			if ($this->options['output_admin_only']) echo "checked='checked'";
		?> />
		Only show the output to logged-in users with Administrator privileges.</label>
		</td>
	</tr>
</table>

<p class="submit"><input type="submit" name="Submit" value="Save Changes" />
</p>
</form>

</div>
<?php
	}
	
	function cmp_total_time($a, $b){
		if ( isset ($a['@meta']['total_time']) ) {
			$x = -($a['@meta']['total_time'] - $b['@meta']['total_time']);
		} else {
			$x = -($a['total_time'] - $b['total_time']);
		}
		return $x<0?-1:($x>0?1:0);
	}
	
	function print_row_with_meta($text, $data, $sort_function = null){
		//Acquire metadata and check if this item has any children
		$has_children = false;
		$addendums = array();
		if ( is_array($data) && isset($data['@meta']) ){
			$meta = $data['@meta'];
			unset($data['@meta']);
			if (count($data) > 0) {
				$has_children = true;
			}
			
			if ( isset($meta['addendums']) )
				$addendums = $meta['addendums'];
		} else {
			$meta = $data;
			$has_children = false;
		}
		
		$has_children = (count($addendums)>0) || $has_children;
		
		//If this item has no subitems then it is on the lowest level
		$subclass = $has_children?'':"wsm_lowest"; 
			
		echo "\t<li class='$subclass'><div class='wsm_mainline $subclass'>\n";

		if (is_array($meta)) {
			if ( isset ($meta['total_time']) )
				echo "<div class='wsm_time'>", sprintf('%.8f', $meta['total_time']),"</div>\n";
			if ( isset ($meta['iterations']) )	
				echo "<div class='wsm_right'>x", $meta['iterations'],"</div>\n";
		} else {
			if ( $meta !== null )
				echo "<div class='wsm_time'>", $meta,"</div>\n";
		}
			
		if ( $has_children )		 
			echo "<a href='javascript:void(0)' class='wsm_collapser'>+</a>  "; 
		echo $this->sanitize_output($text);
		echo "</div>";
			
		if ( $has_children ) {
			$this->print_data_list($data, $addendums, $sort_function);
		}
		echo "</li>\n";
	}
	
	function print_data_list($data, $addendums = null, $sort_function=null, 
		$class='wsm_collapsible', $style='display:none;' 
	) {
		
		//Output subitems. Sort them items if a sort function is provided.
		echo "<ul class='$class' style='$style'>\n";
		if (is_array($addendums)){
			foreach($addendums as $text => $more_text){
				$this->print_row_with_meta($text, $more_text, null);
			}
		}
		if ($sort_function) uasort($data, $sort_function);
		foreach($data as $text => $sub_data){
			$this->print_row_with_meta($text, $sub_data, $sort_function);
		}
		echo "</ul>\n";
	}
	
  /**
   * PluginProfiler::sanitize_output()
   * Applies htmlentities() and replaces non-printable characters with question marks
   *
   * @param string $str
   * @return string
   */
	function sanitize_output($str){
			$str = htmlentities($str);
			$str = preg_replace('/[^\x20-\xF0]/', '?', $str);
			return $str;
	}
	
  /**
   * LoadTimeMonitor::print_advanced_stats()
   * Displays neatly formatted profiler statistics
   *
   * @param array $stats
   * @return void
   */
	function print_advanced_stats(){
		global $plugin_stats, $hook_stats, $query_stats;
		global $plugin_load_stats, $plugin_load_time, $core_load_time;
		
		echo '<style type="text/css">
		.wsm_container {
			/* Very verbose rules to override anything a theme might throw at me */
			width: 580px;
			font-family: "Lucida Grande","Lucida Sans Unicode",Tahoma,Verdana,sans-serif;
			font-style: normal;
			font-variant: normal;
			font-weight: normal;
			font-size: 13px;
			line-height: normal;
			font-size-adjust: none;
			font-stretch: normal;
			background-color: rgb(255, 255, 255);
			color: rgb(51, 51, 51);
			text-align:left;
			padding-right: 40px;
			padding-top: 4px;
			padding-bottom: 4px; 
		}
		
		.wsm_toolbar {
			border: 1px solid #E0E0E0;
			margin: 20px 0px 10px 40px;
			padding: 3px;
		}
		
		.wsm_toolbar a {
			text-decoration:none;
		}
		
		.wsm_container ul {
			padding-left: 40px;
			display: block;
			margin-top: 1em;
			margin-bottom: 1em;
		}
		
		.wsm_container li {
			list-style-type: none;
			border: 1px solid #E0E0E0;
			margin-top: 4px;
			margin-bottom: 4px;
			margin-left: 0px;
			padding: 0px;
			list-style-type: none;
		    color: #333;
		    line-height: 140%;
		}
		
		.wsm_container li  li{
			border-right: none;
		}
		
		.wsm_mainline {
			/* background-color: #f0f0f0; */
			display: block;
			clear: left;
		    padding: 3px 3px 3px 3px;
		    font-size: 85%;
		    position: relative;
		    background: #eaf0f5;
		    /* border-top: 1px solid #c0c7cd; */
		}
		.wsm_time {
			float:right;
			/* border: 1px solid red; */
		}
		.wsm_right {
			float:right;
			margin-right: 4px;
			padding-right: 4px;
			border-right: 1px solid #c0c7cd;
		}
		
		.wsm_lowest, wsm_lowest  li, wsm_lowest wsm_mainline  {
			/* background-color: white !important; */ 
			border: none !important;
		}
		
		.wsm_collapser {
			text-decoration:none;
			font-weight: bold;
			display:block; 
			float:left;
			border: 1px solid #E0E0E0;
			width: 15px;
			height: 15px;
			background: white;
			margin-right: 4px;
			text-align:center;
			line-height: 15px; 
		}
		
		.wsm_collapse_buttons {
			margin-top: 5px;
			font-size: smaller;
		}
		
		.wsm_summary td, .wsm_summary th {
			padding: 4px;
			background-color: #d9e7f2;
		}
		
		.wsm_summary td {
			text-align: center;
			background-color: #ecf2f7;
		}
		
		.wsm_summary th {
			text-align: left;
		}
		
		</style>
		
<script>
$j = jQuery.noConflict();
$j(document).ready(function() {
	$j(".wsm_collapser").click(function() {
		var el = $j(this).parent().parent().children(".wsm_collapsible").get(0);
		//if ( $j(el).visible() ){ //seems to cause problems sometimes
		if ( el.style.display != "none" ){
			el.style.display
			$j(el).hide("fast");
			$j(this).html("+");
		} else {
			$j(el).show("fast");
			$j(this).html("-");
		};
	});
});

var wsm_current_view = "view_by_plugins";

function wsm_expand_all(){
	$j("#"+wsm_current_view+" .wsm_collapsible").show();
	$j("#"+wsm_current_view+" .wsm_collapser").html("-");
}

function wsm_collapse_all(){
	$j("#"+wsm_current_view+" .wsm_collapsible").hide();
	$j("#"+wsm_current_view+" .wsm_collapser").html("+");
}

function wsm_view(view_name){
	$j("#"+wsm_current_view).hide();
	wsm_current_view = view_name;
	$j("#"+view_name).show();
	$j(".wsm_view_switcher a").css("font-weight", "normal");
	$j(".wsm_view_switcher #switcher-"+view_name).css("font-weight", "bold");
}
		
</script>';

		

		echo "<div class='wsm_container'>";
		
		//--- Collect summary data ---
		//Total page generation time, this one's easy
		$total_time = timer_stop(0,5);
		//Now, total hook time(s)...
		$total_hook_time = $core_hook_time = $plugin_hook_time = 0 ;
		foreach ($plugin_stats as $plugin_file => $data){
			$total_hook_time += $data['@meta']['total_time'];
			if ($plugin_file == '[WP Core]'){
				$core_hook_time += $data['@meta']['total_time'];
			} else {
				$plugin_hook_time += $data['@meta']['total_time'];
			}
		}
		
		//Now, total query times...
		$total_query_time = $core_query_time = $plugin_query_time = 0 ;
		foreach($query_stats as $plugin_file => $data){
			$total_query_time += $data['@meta']['total_time'];
			if ($plugin_file == '[WP Core]'){
				$core_query_time += $data['@meta']['total_time'];
			} else {
				$plugin_query_time += $data['@meta']['total_time'];
			}
		}
		
		//Print the summary table
		echo "<div class='wsm_toolbar'>
			<table class='wsm_summary' border='0'>
				<tr>
					<th style='background-color: white;'></th>
					<th>WordPress</th>
					<th>Plugins</th>
					<th>Totals</th>
				</tr>
				<tr>
					<th>Init</th>
					<td>",sprintf('%.5f', $core_load_time),"</td>
					<td>",sprintf('%.5f', $plugin_load_time),"</td>
					<td>",sprintf('%.5f', $plugin_load_time+$core_load_time),"</td>
				</tr>
				<tr>
					<th>Queries</th>
					<td>",sprintf('%.5f', $core_query_time),"</td>
					<td>",sprintf('%.5f', $plugin_query_time),"</td>
					<td>",sprintf('%.5f', $total_query_time),"</td>
				</tr>
				<tr>
					<th>Hooks</th>
					<td>",sprintf('%.5f', $core_hook_time),"</td>
					<td>",sprintf('%.5f', $plugin_hook_time),"</td>
					<td>",sprintf('%.5f', $total_hook_time),"</td>
				</tr>
				<!--
				<tr>
					<th>Other</th>
					<td style='text-align:center'>-</td>
					<td style='text-align:center'>-</td>
					<td>",sprintf('%.5f', $total_time - $total_hook_time - $total_query_time),"</td>
				</tr>
				-->
				<tr>
					<th>Final</th>
					<td></td>
					<td></td>
					<td>",sprintf('%.5f', $total_time),"</td>
				</tr>
			</table>
		</div>";
		
		
		echo "<div class='wsm_toolbar'>
			<div class='wsm_view_switcher'>
				View : 
				<a href='javascript:void(0)' onclick='wsm_view(\"view_by_plugins\")' 
					id='switcher-view_by_plugins'style='font-weight:bold;'>Components</a> | 
				<a href='javascript:void(0)' onclick='wsm_view(\"view_by_hook\")' 
					id='switcher-view_by_hook'>Hooks</a> |
				<a href='javascript:void(0)' onclick='wsm_view(\"view_by_function\")' 
					id='switcher-view_by_function'>Handlers</a> | 
				<a href='javascript:void(0)' onclick='wsm_view(\"view_by_query\")' 
					id='switcher-view_by_query'>Queries</a>
			</div>
			<div class='wsm_collapse_buttons'>
				<a href='javascript:void(0)' onclick='wsm_expand_all()'>Expand All</a> | 
				<a href='javascript:void(0)' onclick='wsm_collapse_all()'>Collapse All</a>
			</div>
		</div>
		";
		
		//View : By Plugin
		$stats_by_plugin = $plugin_stats;
		foreach ($plugin_load_stats as $plugin_file => $time){
			if (isset($stats_by_plugin[$plugin_file])){
				$data = $stats_by_plugin[$plugin_file];
			} else {
				$data = array(
					'@meta' => array(
						'total_time' => 0,
						'iterations' => 0,
						'addendums' => array(),
					),
				);
			}
			
			$data['@meta']['addendums']['Load time'] = array(
					'total_time' => $plugin_load_stats[$plugin_file],
				);
				
			//Note : this is innacurate because various recorded times will overlap. 
			//Anybody willing to fix that?
			$data['@meta']['total_time'] += $plugin_load_stats[$plugin_file];
			
			$stats_by_plugin[$plugin_file] = $data;
		}
		//Enrich the stats by adding info from the query tracker
		foreach ($query_stats as $plugin_file => $stats){
			$data = $stats_by_plugin[$plugin_file];
			
			//Add time spent in queries, but make sure it doesn't overlap with hooks
			$data['@meta']['query_time'] = $stats['@meta']['total_time'];
			$data['@meta']['hooked_time'] = $stats['@meta']['hooked_time'];
			$data['@meta']['total_time'] += ($data['@meta']['query_time'] - $data['@meta']['hooked_time']);
			$data['@meta']['addendums'][$stats['@meta']['iterations']. ' queries'] = array(
					'total_time' => $data['@meta']['query_time']
				);
			
			$stats_by_plugin[$plugin_file] = $data;
		}
		
		echo "<div class='wsm_view' id='view_by_plugins'>";
		$this->print_data_list($stats_by_plugin, null, array($this, 'cmp_total_time'), '', '');
		echo "</div>";
		
		//View : By Hook
		//Must enhance the hook_stats array for this.
		foreach ($stats_by_plugin as $plugin_file => $hooks){
			foreach ($hooks as $tag => $functions){
				if (substr($tag,0,1) == '@') continue; //skip meta
				
				//The below condition is only needed for the shutdown hook, which is when this
				//function itself operates. The data is therefor unreliable and usually unimportant,
				//so I skip it. 
				if ($tag == 'shutdown') continue;
				
				$hook_stats[$tag]['@meta']['plugin_time'] += $functions['@meta']['total_time']; 	
				$hook_stats[$tag][$plugin_file] = $functions; 
			}
		}
		
		foreach ($hook_stats as $tag => &$data){
			if ( count($data) == 1 ){
				$data['@meta']['addendums'] = array('[handled by WordPress core]' => null);
			}
		}
		
		echo "<div class='wsm_view' id='view_by_hook' style='display:none;'>";
		$this->print_data_list($hook_stats, null, array($this, 'cmp_total_time'), '', '');
		echo "</div>";
		
		//View : By Function
		//Must build a new, differently structured array for this.
		$stats_by_function = array();
		foreach ($stats_by_plugin as $plugin_file => $hooks){
			foreach ($hooks as $tag => $functions){
				if (substr($tag,0,1) == '@') continue; //skip meta
				
				foreach($functions as $function => $time){
					if (substr($function,0,1) == '@') continue; //skip meta
					
					//Set up default values if not already done.
					if (!isset($stats_by_function[$function])) 
						$stats_by_function[$function] = array(
							'@meta' => array(
								'total_time' => 0, 
								'plugin_file' => $time['plugin_file'],
								'iterations' => 0,
								'addendums' => array('File : '.$time['plugin_file'] => null),
							)
						 );
					if (!isset($stats_by_function[$function][$tag])) 
						$stats_by_function[$function][$tag] = array(
							'total_time' => 0, 'iterations' => 0
						 );
						
					//Update time totals
					$stats_by_function[$function][$tag]['total_time'] += $time['total_time'];
					$stats_by_function[$function][$tag]['iterations'] += $time['iterations'];
					$stats_by_function[$function]['@meta']['total_time'] += $time['total_time'];
					$stats_by_function[$function]['@meta']['iterations'] += $time['iterations'];
				}
			}
		}
		
		echo "<div class='wsm_view' id='view_by_function' style='display:none;'>";
		$this->print_data_list($stats_by_function, null, array($this, 'cmp_total_time'), '', '');		
		echo "</div>";
		
		echo "<div class='wsm_view' id='view_by_query' style='display:none;'>";
		$this->print_data_list($query_stats, null, array($this, 'cmp_total_time'), '', '');		
		echo "</div>";
		
		//print_r($query_stats);
	}
	
  /**
   * PluginProfiler::print_simple_stats()
   * Uses print_r to output the plugin's global variables.
   *
   * @return void
   */
	function print_simple_stats(){
		global $plugin_stats, $plugin_load_stats, $hook_stats, $query_stats;
		echo "<div align='left'>";
		echo "<br/><strong>Load stats</strong><pre>\n", print_r($plugin_load_stats, true), "\n</pre>";
		echo "<br/><strong>Hook stats</strong><pre>\n", print_r($hook_stats, true), "\n</pre>";
		echo "<strong>Execution stats</strong><pre>\n", print_r($plugin_stats, true), "\n</pre>";
		echo "<br/><strong>Query stats</strong><pre>\n", print_r($query_stats, true), "\n</pre>";
		echo "</div>";
	}
	
}//class declaration

}//class_exists

$plugin_profiler = new PluginProfiler;

?>